#include "box.h"
#include "xyz.h"
#include <math.h>
#include <float.h>
#include <iostream>
#include <fstream>

#undef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#undef min
#define min(a,b) ((a) < (b) ? (a) : (b))

using namespace std;

enum {SS, SB, SWX, SWY, ST}; // collision type : chamber-specific

//enum {DOWN = -1, UP = 1};

Box::~Box()
{
    top_hist_.dump("top.hist");
    top_hist_.dumpZone("top.zone");
    bot_hist_.dump("bot.hist");
}

Box::Box(const char* filename)
    : col_type_(SS)
    , col_st_(0)
    , col_sb_(0)
    , col_swx_(0)
    , col_swy_(0)
{
    ifstream is(filename);
    if(!is.is_open()) {cerr << "Error reading "<< filename <<endl; exit(1);}
    is >> xsize0_ >> ysize0_>> height0_; while(is.get()!='\n');

    is >> base_amplitude_ >> base_period_; while(is.get()!='\n');
    base_period_half_= .5*base_period_; base_period_inv_= 1./base_period_;
    base_velocity_= 4.*base_amplitude_*base_period_inv_;
    base_velocity_inv_= 1./base_velocity_;// if(base_amplitude_)...

    is >> gravity_; while(is.get()!='\n');
    gravity_half_= .5*gravity_; gravity_inv_= 1./gravity_;

    double v_threshold_sb, v_threshold_sw, v_threshold_st;
    is >> coef_sb_ >> v_threshold_sb; while(is.get()!='\n');
    is >> coef_sw_ >> v_threshold_sw; while(is.get()!='\n');
    is >> coef_st_ >> v_threshold_st; while(is.get()!='\n');
    v_scale_sb_= 1./v_threshold_sb;
    v_scale_sw_= 1./v_threshold_sw;
    v_scale_st_= 1./v_threshold_st;
    is >> r_tol_; while(is.get()!='\n');

    is >> base_z_ >> base_dir_; while(is.get()!='\n');

    is >> maps2D_;  while(is.get()!='\n');
    is >> xcells_ >> ycells_; while(is.get()!='\n');
    if(!is) {cerr << "Error reading "<< filename <<endl; exit(1);}

    sprintf(impact_map_top_file_, "top.map");
    sprintf(impact_map_bot_file_, "bot.map");
    if(maps2D_) {
        impact_map_top_.open(impact_map_top_file_);
        impact_map_bot_.open(impact_map_bot_file_);
    }
    double normal[3]= {0., 0., 1.};
    top_hist_.setNormal(normal);
    bot_hist_.setNormal(normal);

}
/*
Box::Box(istream is)
    : col_type_(SS)
    , col_st_(0)
    , col_sb_(0)
    , col_swx_(0)
    , col_swy_(0)
{
    is >> xsize0_ >> ysize0_ >> height0_; while(is.get()!='\n');

    is >> base_amplitude_ >> base_period_; while(is.get()!='\n');
    base_period_half_= .5*base_period_; base_period_inv_= 1./base_period_;
    base_velocity_= 4.*base_amplitude_*base_period_inv_;
    base_velocity_inv_= 1./base_velocity_;// if(base_amplitude_)...

    is >> gravity_; while(is.get()!='\n');
    gravity_half_= .5*gravity_; gravity_inv_= 1./gravity_;

    is >> coef_sw_ >> v_threshold_sw_; while(is.get()!='\n');
    is >> coef_st_ >> v_threshold_st_; while(is.get()!='\n');
    is >> r_tol_; while(is.get()!='\n');

    is >> base_z_ >> base_dir_; while(is.get()!='\n');

    is >> maps2D_;  while(is.get()!='\n');
    is >> xcells_ >> ycells_; while(is.get()!='\n');
}
*/
void
Box::useSphereDiameter(double diam)
{
    height_= height0_ - diam;
    xsize_= xsize0_ - diam*.5;
    ysize_= ysize0_ - diam*.5;
    top_hist_.setCells(xcells_, ycells_, xsize_, ysize_);
    bot_hist_.setCells(xcells_, ycells_, xsize_, ysize_);
}

double
Box::level(double volume) const
{
    return volume / (4*xsize0_*ysize0_);
}

int
Box::isInside(const double *r) const
{
    if(height_>0 && r[Z] > height_ + r_tol_) return 0;
    if(r[Z] < base_z_ - r_tol_) return 0;
    if(fabs(r[X]) > xsize_ || fabs(r[Y]) > ysize_) return 0;
    return 1;
}

int
Box::isOutside(const double *r) const
{
    if(height_>0 && r[Z] > height_ + r_tol_) return 1;
    if(r[Z] < base_z_ - r_tol_) return 1;
    if(fabs(r[X]) > xsize_ || fabs(r[Y]) > ysize_) return 1;
    return 0;
}

void
Box::advance(double goal)// move base
{
    if(base_amplitude_) {
        double t0= (base_amplitude_ - base_dir_*base_z_)*base_velocity_inv_;
        base_z_+= base_dir_ * base_velocity_ *t0;
        while(goal > t0) {
            t0+= base_period_half_;
            base_dir_= -base_dir_;
            base_z_= -base_z_;
        }
        base_z_+= base_dir_ * base_velocity_ * (goal - t0);
    }
}

double
Box::detect(const double *r, const double *v, double best_time)
{
    double t, delta;
    {// DETECT COLLISION WITH BASE
        double tpeak= (base_amplitude_ - base_dir_ * base_z_)*base_velocity_inv_;
        if(base_dir_ < 0) tpeak+= base_period_half_;// next peak of sawtooth motion
        double z_tpeak= r[Z] + tpeak*(v[Z] - tpeak*gravity_half_);
        double dz_tpeak= z_tpeak - base_amplitude_;

        double dz0= r[Z] - base_amplitude_;

        double tentry = 0.;
        if(dz0 > 0. || dz_tpeak > 0.)// the particle does make an entry/reentry later
            tentry= (v[Z] + sqrt(v[Z]*v[Z] + 2.* dz0 * gravity_))*gravity_inv_;

        if(tentry < best_time) {
            tentry-= tpeak;
            tpeak+= base_period_ * ceil(tentry*base_period_inv_);
// the impact is within the cycle preceding "tpeak"
            tpeak-= base_period_half_;
            z_tpeak= r[Z] + tpeak*(v[Z] - tpeak*gravity_half_);
            dz_tpeak= z_tpeak + base_amplitude_;

            double vz_tpeak= v[Z] - tpeak*gravity_;
            double dvz_tpeak;
            if(dz_tpeak < 0 && vz_tpeak < 0) dvz_tpeak= vz_tpeak + base_velocity_;
            else                             dvz_tpeak= vz_tpeak - base_velocity_;
            delta = dvz_tpeak*dvz_tpeak + 2.* dz_tpeak * gravity_;
            t= tpeak + (dvz_tpeak + sqrt(delta))*gravity_inv_; // DIVISION!

            if(t < best_time) {best_time= t; col_type_= SB;}
        }
    }

    {// DETECT COLLISION WITH WALLS
        double dx, dy, tx, ty;
        if(v[X]>0) dx= xsize_-r[X]; else dx=-xsize_-r[X]; tx= dx/v[X];
        if(v[Y]>0) dy= ysize_-r[Y]; else dy=-ysize_-r[Y]; ty= dy/v[Y];
        if(tx < best_time) {best_time= tx; col_type_= SWX;}
        if(ty < best_time) {best_time= ty; col_type_= SWY;}
    }

    if(height_>0 && v[Z]>0.) {// DETECT COLLISION WITH TOP
        t= DBL_MAX;
        double vz2= v[Z]*v[Z];
        delta= vz2 - 2.*(height_ - r[Z])*gravity_;
        if(delta > 0.) t= (v[Z] - sqrt(delta))*gravity_inv_;
        if(t < best_time) {best_time= t; col_type_= ST;}
    }
    return best_time;
}

void
Box::collide(const double *r, double *v)
{
    double v_new[3];
    double vz_new;
    switch(col_type_) {
        case SB :{
            ++col_sb_;
            double base_velocity= base_velocity_*base_dir_;
            double vz_impact= base_velocity - v[Z];
            vz_new= base_velocity + coef_sb(vz_impact)*vz_impact;
//            vz_new= 1.2 * base_velocity_*base_dir_ - .2*v[Z];

            if(maps2D_) dumpImpact2D(r, impact_map_bot_);

            v_new[X]=v[X]; v_new[Y]=v[Y]; v_new[Z]= vz_new;
            bot_hist_.hit(r, v, v_new);
            v[Z]= vz_new;
            }
            break;
        case SWX :{
                ++col_swx_;
                v[X]*= -coef_sw(fabs(v[X]));
            }
            break;
        case SWY :{
                ++col_swy_;
                v[Y]*= -coef_sw(fabs(v[Y]));
            }
            break;
        case ST :
            ++col_st_;
            vz_new= -coef_st(v[Z])*v[Z];

            if(maps2D_) dumpImpact2D(r, impact_map_top_);

            v_new[X]=v[X]; v_new[Y]=v[Y]; v_new[Z]= vz_new;
            top_hist_.hit(r, v, v_new);
            v[Z]= vz_new;
            break;
    }
}

double
Box::coef_sb(double velocity) const
{
    velocity*= v_scale_sb_;
    if(velocity < 1.) return coef_sb_;// elastic
    return coef_sb_*pow(velocity, -0.25);// v^-0.25 decay
}

double
Box::coef_sw(double velocity) const
{
    velocity*= v_scale_sw_;
    if(velocity < 1.) return coef_sw_;// elastic
    return coef_sw_*pow(velocity, -0.25);// v^-0.25 decay
}

double
Box::coef_st(double velocity) const
{
    velocity*= v_scale_st_;
    if(velocity < 1.) return coef_st_;// elastic
    return coef_st_*pow(velocity, -0.25);// v^-0.25 decay
}

void
Box::dumpImpact2D(const double* r, const char* filename) const
{
    ofstream os;
    os.open(filename, ofstream::app);
    os << r[X] <<" "<<r[Y]<< endl;
}

void
Box::dumpImpact2D(const double* r, ofstream& os) const
{
    os << r[X] <<" "<<r[Y]<< endl;
}

void
Box::reset()
{
    col_type_= SS;
    col_st_= 0;
    col_sb_= 0;
    col_swx_= 0;
    col_swy_= 0;

    if(impact_map_top_.is_open()) impact_map_top_.close();
    if(impact_map_bot_.is_open()) impact_map_bot_.close();

    if(maps2D_) {
        impact_map_top_.open(impact_map_top_file_);
        impact_map_bot_.open(impact_map_bot_file_);
    }

    top_hist_.dump("top.hist"); top_hist_.reset();
    bot_hist_.dump("bot.hist"); bot_hist_.reset();
}
