#include "shot.h"
#include "chamber.h"
#include "stats3D.h"
#include "xyz.h"
#include "echo.h"
#include <math.h>
#include <float.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>

using namespace std;

Shot::~Shot()
{
    char filename[]= "state.out";
    ofstream os(filename);
    if(!os.is_open()) {cerr <<"Error writing "<< filename << endl; exit(1);}
    dumpState(os);
    if(!os) {cerr <<"Error writing "<< filename << endl; exit(1);}

// free state vector
    int sphere;
    delete[] r_[0]; delete[] r_;
    delete[] v_[0]; delete[] v_;

// free collision data
    delete[] tnext_; delete[] partner_; delete[] update_;
}

Shot::Shot(const char* filename)
    : r_(NULL)
    , v_(NULL)
    , tnext_(NULL)
    , partner_(NULL)
    , update_(NULL)
{
    ifstream is(filename);
    if(!is.is_open()) {cerr <<"Error reading "<< filename << endl; exit(1);}

// constants
    is >> n_ >> diam_; while(is.get()!='\n');
    diam2_= diam_*diam_; radius_= .5*diam_;

    is >> eta_ >> kt_ >> omega0_; while(is.get()!='\n');

    double v_threshold;
    is >> coef0_ >> v_threshold; while(is.get()!='\n');
    v2_scale_= 1./(v_threshold*v_threshold);
    
    is >> rtol_; while(is.get()!='\n');
    r2tol_= 2*diam_*rtol_;

    if(!is) {cerr << "Error reading " << filename << endl; exit(1);}
    is.close();

    tnext_= new double[n_];// time to next collision for each particle
    partner_= new int[n_];// its collision partner
    update_ = new int[n_];// collision update flag

    r_= new double*[n_]; v_= new double*[n_];// state vector
    r_[0]= new double[3*n_]; v_[0]= new double[3*n_];
    for(int sphere=1; sphere<n_; ++sphere) {
        r_[sphere]= r_[sphere-1] + 3;
        v_[sphere]= v_[sphere-1] + 3;
    }
}

void
Shot::set(Chamber& chamber) const
{
    chamber.useSphereDiameter(diam_);
}

long
Shot::distribute(Chamber& chamber)
{
    int good;
    long seed= time(0); srand(seed);
    double r_max= chamber.radius();
    double volume= n_ * M_PI * diam2_ * diam_ /(6. * eta_);
    double z_max= chamber.level(volume);

    cout <<"Distributing shot randomly... ";
    int i=0;
    while(i<n_) {
        double *ri= r_[i];
        do {
            ri[X]= (2.* rand()/double(RAND_MAX) - 1.)*r_max;
            ri[Y]= (2.* rand()/double(RAND_MAX) - 1.)*r_max;
            ri[Z]= rand()/double(RAND_MAX) * z_max;
        } while(chamber.isOutside(ri));
        good= 1;
        for(int j=0; j<i; ++j) {
            double *rj= r_[j];
            double dx= ri[X] - rj[X];
            double dy= ri[Y] - rj[Y];
            double dz= ri[Z] - rj[Z];
            double dr2= dx*dx + dy*dy + dz*dz;
            if(dr2 < diam2_) good= 0;
        }
        if(good) ++i;
    }
    cout <<"DONE"<< endl;

// pseudo-normal velocity distribution
    double svx = 0., svy = 0., svz = 0.;

    for(i=0; i<n_; ++i) {
        double *vi= v_[i];
        vi[X]= 2.*rand()/double(RAND_MAX) - 1.; svx+= vi[X];
        vi[Y]= 2.*rand()/double(RAND_MAX) - 1.; svy+= vi[Y];
        vi[Z]= 2.*rand()/double(RAND_MAX) - 1.; svz+= vi[Z];
    }
    svx/= n_; svy/= n_; svz/= n_;

    if(n_==1) svx= svy= svz= 0;

    double svx2= 0., svy2= 0., svz2= 0.;
    for(i=0; i<n_; ++i) {
        double *vi= v_[i];
        vi[X]-= svx; svx2+= vi[X]*vi[X];
        vi[Y]-= svy; svy2+= vi[Y]*vi[Y];
        vi[Z]-= svz; svz2+= vi[Z]*vi[Z];
    }

    double alpha= sqrt(n_*kt_ * 3. / (svx2 + svy2 + svz2));
    double mom_inertia= 0;// moment of inertia
    double mom_cinetic= 0;// cinetic moment
    for(i=0; i<n_; ++i) {
        double *ri= r_[i], *vi= v_[i];
        mom_inertia+= ri[X]*ri[X] + ri[Y]*ri[Y];
        vi[X]*= alpha; vi[Y]*= alpha; vi[Z]*= alpha;
        mom_cinetic+= ri[X]*vi[Y] - ri[Y]*vi[X];
    }
    double omega= mom_cinetic/mom_inertia;// solid rotation
    for(i=0; i<n_; ++i) {
        double *ri= r_[i], *vi= v_[i];
        vi[X]-= (omega0_ - omega)*ri[Y];
        vi[Y]+= (omega0_ - omega)*ri[X];
    }
    resetCollisionData();
    return seed;
}

void
Shot::detect()
{
    for(int i=0; i<n_; ++i) if(update_[i]) {
        double *ri= r_[i], xi = ri[X], yi = ri[Y], zi = ri[Z];
        double *vi= v_[i], vxi= vi[X], vyi= vi[Y], vzi= vi[Z];

        for(int j=0; j<n_; ++j) if(!update_[j] || j<i) {// for non-redundancy
            double *rj= r_[j],  dx=  xi - rj[X],  dy=  yi - rj[Y],  dz=  zi - rj[Z];
            double *vj= v_[j], dvx= vxi - vj[X], dvy= vyi - vj[Y], dvz= vzi - vj[Z];

            double b= dx*dvx + dy*dvy + dz*dvz;
            if(b < 0.) {// the mutual distance is decreasing

// equation to solve in t :
// (dr + t*dv)^2 = sigma^2 : a*t^2 + 2*b*t + c = 0
// with a = dv*dv; b = dr*dv; c = dr*dr - sigma^2

                double a= dvx*dvx + dvy*dvy + dvz*dvz;
                double c= dx*dx + dy*dy + dz*dz - diam2_;
/*
                if(c < -r2tol_)
                    cout <<"negative mutual distance "<< c <<endl
                         <<"between particles "<< i <<" and "<< j <<endl;
*/
                double delta= b*b - a*c;
                if(delta > 0.) {
                    double t= -(b + sqrt(delta))/a; // we need the SMALLER solution
/*
                    if(t < 0.) // this happens only if c is negative
                        cout <<"negative sphere-sphere impact time "<< t <<endl
                             <<"for particles "<< i <<" and "<< j <<endl;
*/
                    if(t < tnext_[i]) { // set next collision for i
                        tnext_[i]= t; partner_[i]= j;
                    }
                    if(t < tnext_[j]) { // set next collision for j
                        tnext_[j]= t; partner_[j]= i;
                    }
                }
            }
        }
    }
    best_time_= DBL_MAX;
    for(int sphere=0; sphere<n_; ++sphere) if (tnext_[sphere] < best_time_) {
        best_time_= tnext_[sphere];
        chosen_= sphere;
    }
    if(best_time_<0.) cout <<"NEGATIVE INTRASHOT IMPACT TIME : "<<best_time_<<endl;
}

void
Shot::detect(Chamber& chamber)
{
    chamber_impact_= 0;
    for(int sphere=0; sphere<n_; ++sphere) {
        double newtime= chamber.detect(r_[sphere], v_[sphere], best_time_);
        if(newtime < best_time_) {
            chosen_= sphere;
            best_time_= newtime;
            chamber_impact_= 1;
        }
    }
}

void
Shot::advance(double gravity, double time)
{
    double vshift= -time*gravity, zshift= .5*time*vshift;
    for(int i=0; i<n_; ++i) {
        double *ri= r_[i], *vi= v_[i];
        ri[X]+= time * vi[X];
        ri[Y]+= time * vi[Y];
        ri[Z]+= time * vi[Z] + zshift;
        vi[Z]+= vshift;
        tnext_[i]-= time;
    }
    best_time_-= time;
}

void
Shot::collide(Chamber& chamber)
{
    double *r_chosen= r_[chosen_], *v_chosen= v_[chosen_];

    if(chamber_impact_) {
        chamber.collide(r_chosen, v_chosen);
        for(int sphere=0; sphere< n_; ++sphere) {
            update_[sphere]= 0;
            if(chosen_ == sphere || chosen_ == partner_[sphere]) {
                update_[sphere]= 1;
                tnext_[sphere]= DBL_MAX;
                partner_[sphere]= -1;
            }
        }
    }
    else {
        ++collision_;
        int partner= partner_[chosen_];
        double *r_partner= r_[partner], *v_partner= v_[partner];
/*
        if(partner < 0 || partner >= n_) {
            cout << "#ERROR sphere-sphere collision number "<<collision_ << endl
                 << "#partner of sphere "<<chosen_<<" has index "<<partner<<" out of range"
                 << endl;
            exit(1);
        }
*/
        double
            dx= r_chosen[X] - r_partner[X], dy= r_chosen[Y] - r_partner[Y],
            dz= r_chosen[Z] - r_partner[Z], dr2= dx*dx + dy*dy + dz*dz;
/*
        if(fabs(dr2 - diam2_)> r2tol_) {
            cout << "#WARNING sphere-sphere collision number "<<collision_ <<endl
                 << "#spheres "<<chosen_<<" and "<<partner<<" are not in contact"<<endl;
        }
*/
        double
            dvx= v_chosen[X] - v_partner[X], dvy= v_chosen[Y] - v_partner[Y],
            dvz= v_chosen[Z] - v_partner[Z], dv2= dvx*dvx + dvy*dvy + dvz*dvz;

        double renorm= .5*(1. + coef(dv2))*(dx*dvx + dy*dvy + dz*dvz)/diam2_;
        dx*= renorm; dy*= renorm; dz*= renorm;
        v_chosen[X]-= dx; v_partner[X]+= dx;
        v_chosen[Y]-= dy; v_partner[Y]+= dy;
        v_chosen[Z]-= dz; v_partner[Z]+= dz;
        
        for(int sphere=0; sphere< n_; ++sphere){
            update_[sphere]= 0;
            if(partner == sphere
            || chosen_ == partner_[sphere]
            || partner == partner_[sphere]) {
                update_[sphere]= 1;
                tnext_[sphere]= DBL_MAX;
                partner_[sphere]= -1;
            }
        }
    }
}

void
Shot::test(const Chamber& chamber) const
{
    cout <<"Testing for overlaps..."<< endl;
    for(int i=0; i<n_; ++i) {
        double *ri= r_[i];
        if(chamber.isOutside(ri)) {cout <<"NOT IN CHAMBER"<< endl; exit(1);}
        for(int j=0; j <i; ++j) {
            double *rj= r_[j];
            double
                dx= ri[X] - rj[X], dy= ri[Y] - rj[Y], dz= ri[Z] - rj[Z],
                dr2= dx*dx + dy*dy + dz*dz;

            if(dr2 < diam2_ - r2tol_) {cout <<"OVERLAP"<< endl; exit(1);}
        }
    }
}

void
Shot::readState(istream& is)
{
    cout <<"Reading state vector from file..."<< endl;
    for(int i=0; i< n_; ++i) {
        double *ri= r_[i], *vi= v_[i];
        is >> ri[X] >> ri[Y] >> ri[Z] >> vi[X] >> vi[Y] >> vi[Z];
    }
    resetCollisionData();
}

void
Shot::dumpState(ostream& os) const
{
    cout <<"Dumping state vector to file... ";
    for(int i=0; i< n_; ++i) {
        double *ri= r_[i], *vi= v_[i];
        os << ri[X] <<" "<< ri[Y] <<" "<< ri[Z] <<" "
           << vi[X] <<" "<< vi[Y] <<" "<< vi[Z] <<endl;
    }
    cout <<"DONE"<< endl;
}
/*
void
Shot::dumpState(Stats3D& stats) const
{
    stats.newEntry();
    for(int i=0; i< n_; ++i) stats.hit(r_[i], v_[i]);
}

void
Shot::makeStatEntry(Chamber& chamber) const
{
    chamber.stat3DnewEntry();
    for(int i=0; i<n_; ++i) chamber.stat3Dhit(r_[i], v_[i]);
}
*/
double
Shot::coef(double velocity2) const
{
// v_scale_ is the inverse of the threshold velocity squared
    if(velocity2 * v2_scale_ < 1.) return coef0_;// elastic
    return coef0_*pow(velocity2 * v2_scale_, -0.125);// v^-0.25 decay
}

void
Shot::resetCollisionData()
{
    collision_= 0;
    for(int sphere=0; sphere<n_; ++sphere) {
        tnext_[sphere] = 1e8;
        partner_[sphere]= -1;
        update_[sphere]= 1;
    }
    best_time_= DBL_MAX;
    chosen_= -1;
}


