#include "stats.h"
#include "xyz.h"
#include <iostream>
#include <fstream>

using namespace std;

// Stats2Dr
Stats2Dr::~Stats2Dr()
{
    delete[] hist_;
    delete[] pressure_;
    delete[] energy_;
}

Stats2Dr::Stats2Dr()
    : cells_(0)
    , hits_(0)
    , hist_(NULL)
    , pressure_(NULL)
    , energy_(NULL)
{
}

Stats2Dr::Stats2Dr(const double* normal)
    : cells_(0)
    , hits_(0)
    , hist_(NULL)
    , pressure_(NULL)
    , energy_(NULL)
{
    setNormal(normal);
}

void
Stats2Dr::setCells(int cells, double radius)
{
    delete[] hist_;
    cells_= cells;
    hits_= 0;
    scale_= cells_/radius;
    scale_inv_= radius/cells_;

    hist_= new int[cells_];
    pressure_= new double[cells_];
    energy_= new double[cells_];
    for(int cell= 0; cell<cells_; ++cell) {
        hist_[cell]= 0;
        pressure_[cell]= 0.;
        energy_[cell]= 0.;
    }
}

void
Stats2Dr::setNormal(const double* normal)
{
    normal_[X]= normal[X];
    normal_[Y]= normal[Y];
    normal_[Z]= normal[Z];
}

void
Stats2Dr::reset()
{
    hits_= 0;
    for(int cell= 0; cell<cells_; ++cell) {
        hist_[cell]= 0;
        pressure_[cell]= 0.;
        energy_[cell]= 0.;
    }
}

void
Stats2Dr::hit(const double* r, const double* v_before, const double* v_after)
{
    ++hits_;

    double xtemp= normal_[Y]*r[Z]-normal_[Z]*r[Y];
    double ytemp= normal_[Z]*r[X]-normal_[X]*r[Z];
    double ztemp= normal_[X]*r[Y]-normal_[Y]*r[X];
    double radius= sqrt(xtemp*xtemp + ytemp*ytemp + ztemp*ztemp);

    int cell= int(radius*scale_);
    if(cell>-1 && cell<cells_) {
        ++hist_[cell];
        double vn_before= v_before[X]*normal_[X]
                        + v_before[Y]*normal_[Y]
                        + v_before[Z]*normal_[Z];
        double vn_after= v_after[X]*normal_[X]
                        + v_after[Y]*normal_[Y]
                        + v_after[Z]*normal_[Z];
        pressure_[cell]+= vn_before-vn_after;
        energy_[cell]+= vn_before*vn_before-vn_after*vn_after;
    }
}

void
Stats2Dr::dump(const char* filename) const
{
    ofstream os(filename);
    dump(os);
}

void
Stats2Dr::dump(ofstream& os) const
{
    double sum= 0, psum= 0, esum= 0;
    for(int cell=0; cell<cells_; ++cell) {
        double radius_inv= 1./((cell + .5)*scale_inv_);
        sum+= hist_[cell]*radius_inv;
        psum+= pressure_[cell]*radius_inv;
        esum+= energy_[cell]*radius_inv;
    }
    sum= cells_/sum;
    psum= cells_/psum;
    esum= cells_/esum;

    os <<"# "<< hits_ <<" hits"<< endl;
    os <<"# "<< 1 <<' '<<cells_<<' '<<0<<" block, lines, blanks"<< endl;
    os <<"# "<< 3 <<" columns (density, pressure, energy)"<< endl;

    for(int cell=0; cell<cells_; ++cell) {
        double radius= (cell + .5)*scale_inv_;
        double radius_inv= 1./radius;
        os << radius << " "
           << hist_[cell]*sum*radius_inv << " "
           << pressure_[cell]*psum*radius_inv << " "
           << energy_[cell]*esum*radius_inv << endl;
    }
}

void
Stats2Dr::rescale()
{
}

// Stats2Dxy
Stats2Dxy::~Stats2Dxy()
{
    delete[] hist_;
    delete[] pressure_;
    delete[] energy_;
}

Stats2Dxy::Stats2Dxy()
    : xcells_(0)
    , ycells_(0)
    , cells_(0)
    , hits_(0)
    , hist_(NULL)
    , pressure_(NULL)
    , energy_(NULL)
{
}

Stats2Dxy::Stats2Dxy(const double* normal)
    : xcells_(0)
    , ycells_(0)
    , cells_(0)
    , hits_(0)
    , hist_(NULL)
    , pressure_(NULL)
    , energy_(NULL)
{
    setNormal(normal);
}

void
Stats2Dxy::setCells(int xcells, int ycells, double xsize, double ysize)
{
    delete[] hist_;
    xcells_= xcells; ycells_= ycells;
    xscale_= xcells_/xsize; yscale_= ycells_/ysize;
    xscale_inv_= xsize/xcells_; yscale_inv_= ysize/ycells_;

    cells_= xcells_*ycells_;
    hist_= new int[cells_];
    pressure_= new double[cells_];
    energy_= new double[cells_];
    for(int cell= 0; cell<cells_; ++cell) {
        hist_[cell]= 0;
        pressure_[cell]=0;
        energy_[cell]=0;
    }

    hits_= 0;
}

void
Stats2Dxy::setNormal(const double* normal)
{
    normal_[X]= normal[X];
    normal_[Y]= normal[Y];
    normal_[Z]= normal[Z];
}

void
Stats2Dxy::reset()
{
    for(int cell= 0; cell<cells_; ++cell) {
        hist_[cell]= 0;
        pressure_[cell]=0;
        energy_[cell]=0;
    }
    hits_= 0;
}

void
Stats2Dxy::hit(const double* r, const double* v_before, const double* v_after)
{
    ++hits_;

    int xcell= int(.5*(r[X]*xscale_ + xcells_));// TO BE GENERALISED !!!
    int ycell= int(.5*(r[Y]*yscale_ + ycells_));// TO BE GENERALISED !!!
// IDEA : make a "center" vector along with "normal" vector
// "x" axis should be chosen horizontal and orthogonal to "center"
// or identical to regular "X" if "center" is vertical.

    int cell= ycell + xcells_*xcell;
    if(cell>-1 && cell<cells_) {
        ++hist_[cell];
        double vn_before= v_before[X]*normal_[X]
                        + v_before[Y]*normal_[Y]
                        + v_before[Z]*normal_[Z];
        double vn_after= v_after[X]*normal_[X]
                        + v_after[Y]*normal_[Y]
                        + v_after[Z]*normal_[Z];
        pressure_[cell]+= vn_before - vn_after;
        energy_[cell]+= vn_before*vn_before - vn_after*vn_after;
    }
}

void
Stats2Dxy::dump(const char* filename) const
{
    ofstream os(filename);
    if(!os.is_open()) {cerr << "Error writing "<< filename << endl; exit(1);}
    dump(os);
}

void
Stats2Dxy::dumpZone(const char* filename) const
{
    ofstream os(filename);
    if(!os.is_open()) {cerr << "Error writing "<< filename << endl; exit(1);}
    dumpZone(os);
}


void
Stats2Dxy::dump(ofstream& os) const
{
    double sum= 0, psum= 0, esum= 0;
    for(int cell=0; cell<cells_; ++cell) {
        sum+= hist_[cell];
        psum+= pressure_[cell];
        esum+= energy_[cell];
    }
    sum= cells_/sum;
    psum= cells_/psum;
    esum= cells_/esum;
    for(int xcell=0; xcell<xcells_; ++xcell) {
        for(int ycell=0; ycell<ycells_; ++ycell) {
            int cell= ycell + xcells_*xcell;
            os << (2.*xcell - xcells_ + 1.)*xscale_inv_ << " "
               << (2.*ycell - ycells_ + 1.)*yscale_inv_ << " "
               << hist_[cell]*sum << " "
               << pressure_[cell]*psum << " "
               << energy_[cell]*esum << endl;
        }
        os << endl;
    }
}

void
Stats2Dxy::dumpZone(ofstream& os) const
{
    double sum= 0, psum= 0, esum= 0;
    for(int cell=0; cell<cells_; ++cell) {
        sum+= hist_[cell];
        psum+= pressure_[cell];
        esum+= energy_[cell];
    }
    sum= cells_/sum;
    psum= cells_/psum;
    esum= cells_/esum;

// dump Gamma-X section (4 or 2 samples)
    int ycell1= ycells_/2;
    int ycell2= ycells_-1 - ycell1;
    for(int xcell= 0; xcell<(xcells_+1)/2; ++xcell) {
        int cell1a= ycell1 + xcells_*xcell;
        int cell1b= ycell1 + xcells_*(xcells_-1-xcell);
        int cell2a= ycell2 + xcells_*xcell;
        int cell2b= ycell2 + xcells_*(xcells_-1-xcell);

        os << (xcells_ - 2.*xcell - 1.)*xscale_inv_ << " "
           << 0 << " "
           << .25*(hist_[cell1a]+hist_[cell1b]+hist_[cell2a]+hist_[cell2b])*sum << " "
           << .25*(pressure_[cell1a]+pressure_[cell1b]+pressure_[cell2a]+pressure_[cell2b])*psum << " "
           << .25*(energy_[cell1a]+energy_[cell1b]+energy_[cell2a]+energy_[cell2b])*esum << endl;
    }
    os << endl;
    os << endl;

// dump X-M section (4 samples)
    int xcell1= 0;
    int xcell2= xcells_-1;
    for(int ycell=0; ycell<(ycells_+1)/2; ++ycell) {
        int cell1a= ycell + xcells_*xcell1;
        int cell1b= (ycells_ - 1 - ycell) + xcells_*xcell1;
        int cell2a= ycell + xcells_*xcell2;
        int cell2b= (ycells_ - 1 - ycell) + xcells_*xcell2;

//cell1b= cell1a;cell2a= cell1a;cell2b= cell1a;
        os << (xcells_ - 1.)*xscale_inv_ << " "
           << (ycells_ -2.*ycell - 1.)*yscale_inv_  << " "
           << .25*(hist_[cell1a]+hist_[cell1b]+hist_[cell2a]+hist_[cell2b])*sum << " "
           << .25*(pressure_[cell1a]+pressure_[cell1b]+pressure_[cell2a]+pressure_[cell2b])*psum << " "
           << .25*(energy_[cell1a]+energy_[cell1b]+energy_[cell2a]+energy_[cell2b])*esum << endl;
    }
    os << endl;
    os << endl;

// dump M-Gamma section (4 samples)
    if(xcells_<=ycells_) for(int xcell= 0; xcell<(xcells_+1)/2; ++xcell) {
        int cell1a= xcell + xcells_*xcell;
        int cell1b= xcell + xcells_*(xcells_-1-xcell);
        int cell2a= (xcells_-1-xcell) + xcells_*xcell;
        int cell2b= (xcells_-1-xcell) + xcells_*(xcells_-1-xcell);

        os << (xcells_ - 2.*xcell - 1.)*xscale_inv_ << " "
           << (xcells_ - 2.*xcell - 1.)*xscale_inv_ << " "
           << .25*(hist_[cell1a]+hist_[cell1b]+hist_[cell2a]+hist_[cell2b])*sum << " "
           << .25*(pressure_[cell1a]+pressure_[cell1b]+pressure_[cell2a]+pressure_[cell2b])*psum << " "
           << .25*(energy_[cell1a]+energy_[cell1b]+energy_[cell2a]+energy_[cell2b])*esum << endl;
    }
    os << endl;
    os << endl;
}
