#include "stats3D.h"
#include "shot.h"
#include "xyz.h"
#include <iostream>
#include <fstream>

using namespace std;

//------------------------------------------------------------------------------
Stats3Drz::~Stats3Drz()
{
    delete[] n_;
    delete[] vr_; delete[] vt_; delete[] vz_;
    delete[] kr_; delete[] kt_; delete[] kz_;
}

Stats3Drz::Stats3Drz(const char* filename)
    : entries_(0)
    , n_(NULL)
    , vr_(NULL), vz_(NULL), vt_(NULL)
    , kr_(NULL), kz_(NULL), kt_(NULL)
{
    {
        ifstream is(filename);
        if(!is.is_open()) {cerr << "Error reading "<< filename << endl; exit(1);}
        double radius; is >> rcells_ >> radius; while(is.get()!='\n');
        rscale_= rcells_/radius; rscale_inv_= radius/rcells_;
        double height; is >> zcells_ >> height; while(is.get()!='\n');
        zscale_= zcells_/height; zscale_inv_= height/zcells_;
        if(!is) {cerr << "Error reading "<< filename << endl; exit(1);}
    }
    cells_= rcells_*zcells_;
    n_= new int[cells_];
    vr_= new double[cells_]; vz_= new double[cells_]; vt_= new double[cells_];
    kr_= new double[cells_]; kz_= new double[cells_]; kt_= new double[cells_];
    for(int cell= 0; cell<cells_; ++cell) {
        n_[cell]= 0;
        vr_[cell]= 0.; vz_[cell]= 0.; vt_[cell]= 0.;
        kr_[cell]= 0.; kz_[cell]= 0.; kt_[cell]= 0.;
    }
}

void
Stats3Drz::reset()
{
    for(int cell= 0; cell<cells_; ++cell) {
        n_[cell]= 0;
        vr_[cell]= 0.; vz_[cell]= 0.; vt_[cell]= 0.;
        kr_[cell]= 0.; kz_[cell]= 0.; kt_[cell]= 0.;
    }
    entries_= 0;
}

void
Stats3Drz::makeEntry(const Shot& shot)
{
    ++entries_;
    int chunk_size= 3*shot.n();
    const double *r0= shot.r_chunk();
    const double *v0= shot.v_chunk();
    for(int i= 0; i<chunk_size; i+=3) {
        double x = (r0 + i)[X], y = (r0 + i)[Y], z = (r0 + i)[Z];
        double vx= (v0 + i)[X], vy= (v0 + i)[Y], vz= (v0 + i)[Z];
        double r= sqrt(x*x + y*y);
        double radius_inv= 1/r;
        int rcell= int(r*rscale_);
        int zcell= int(z*zscale_);
        int cell= rcell + rcells_*zcell;
        double vr= (vx*x + vy*y)*radius_inv;
        double vt= (vy*x - vx*y)*radius_inv;
        if(cell>-1 && cell<cells_) {
            ++n_[cell];
            vr_[cell]+= vr; vz_[cell]+= vz; vt_[cell]+= vt;
            kr_[cell]+= vr*vr; kz_[cell]+= vz*vz; kt_[cell]+= vt*vt;
        }
    }
}

void
Stats3Drz::dump(const Shot& shot) const
{
// OUTPUT FILE PREPARATION
    const char raw_r_[]= "raw.r";
    const char raw_rz_[]= "raw.rz";
    const char all_r_[]= "all.r";
    const char all_rz_[]= "all.rz";
    const char rho_eta_r_[]= "rho_eta.r";
    const char rho_eta_rz_[]= "rho_eta.rz";
    const char vr_vt_vz_rz_[]= "vr_vt_vz.rz";
    const char temperature_r_[]= "temperature.r";
    const char temperature_rz_[]= "temperature.rz";

    ofstream raw_r(raw_r_);
    ofstream raw_rz(raw_rz_);
    ofstream all_r(all_r_);
    ofstream all_rz(all_rz_);
    ofstream rho_eta_r(rho_eta_r_);
    ofstream rho_eta_rz(rho_eta_rz_);
    ofstream vr_vt_vz_rz(vr_vt_vz_rz_);
    ofstream temperature_r(temperature_r_);
    ofstream temperature_rz(temperature_rz_);
    if(!raw_r.is_open()) {
        cerr << "Error writing "<< all_r_ << endl;
        exit(1);
    }
    if(!raw_rz.is_open()) {
        cerr << "Error writing "<< all_rz_ << endl;
        exit(1);
    }
    if(!all_r.is_open()) {
        cerr << "Error writing "<< all_r_ << endl;
        exit(1);
    }
    if(!all_rz.is_open()) {
        cerr << "Error writing "<< all_rz_ << endl;
        exit(1);
    }
    if(!rho_eta_r.is_open()) {
        cerr << "Error writing "<< rho_eta_r_ << endl;
        exit(1);
    }
    if(!rho_eta_rz.is_open()) {
        cerr << "Error writing "<< rho_eta_rz_ << endl;
        exit(1);
    }
    if(!vr_vt_vz_rz.is_open()) {
        cerr << "Error writing "<< vr_vt_vz_rz_ << endl;
        exit(1);
    }
    if(!temperature_r.is_open()) {
        cerr << "Error writing "<< temperature_r_ << endl;
        exit(1);
    }
    if(!temperature_rz.is_open()) {
        cerr << "Error writing "<< temperature_rz_ << endl;
        exit(1);
    }
    raw_r <<"# "<< entries_ <<" entries"<< endl;
    raw_rz <<"# "<< entries_ <<" entries"<< endl;
    all_r <<"# "<< entries_ <<" entries"<< endl;
    all_rz <<"# "<< entries_ <<" entries"<< endl;
    rho_eta_r <<"# "<< entries_ <<" entries"<<endl;
    rho_eta_rz <<"# "<< entries_ <<" entries"<< endl;
    vr_vt_vz_rz <<"# "<< entries_ <<" entries"<< endl;
    temperature_r <<"# "<< entries_ <<" entries"<< endl;
    temperature_rz <<"# "<< entries_ <<" entries"<< endl;

    raw_r <<"# "<< zcells_ <<' '<< rcells_ <<' '<< 2 <<" zcells, rcells, blanks"<< endl;
    raw_rz <<"# "<< zcells_ <<' '<< rcells_ <<' '<< 1 <<" zcells, rcells, blanks"<< endl;
    all_r <<"# "<< zcells_ <<' '<< rcells_ <<' '<< 2 <<" zcells, rcells, blanks"<< endl;
    all_rz <<"# "<< zcells_ <<' '<< rcells_ <<' '<< 1 <<" zcells, rcells, blanks"<< endl;
    rho_eta_r <<"# "<< zcells_ <<' '<< rcells_ <<' '<< 2 <<" zcells, rcells, blanks"<< endl;
    rho_eta_rz <<"# "<< zcells_ <<' '<< rcells_ <<' '<< 1 <<" zcells, rcells, blanks"<< endl;
    vr_vt_vz_rz <<"# "<< zcells_ <<' '<< rcells_ <<' '<< 1 <<" zcells, rcells, blanks"<< endl;
    temperature_r <<"# "<< zcells_ <<' '<< rcells_ <<' '<< 2 <<" zcells, rcells, blanks"<< endl;
    temperature_rz <<"# "<< zcells_ <<' '<< rcells_ <<' '<< 1 <<" zcells, rcells, blanks"<< endl;

    raw_r <<"# "<< 7 <<" columns (n, vr, vt, vz, kr, kt, kz)"<< endl;
    raw_rz <<"# "<< 7 <<" columns (n, vr, vt, vz, kr, kt, kz)"<< endl;
    all_r <<"# "<< 16 <<" columns (n,rho,eta,n*vr,n*vt,n*vz,n*kr,n*kt,n*kz,vr,vt,vz,kr-vr^2,kt-vt^2,kz-vz^2,T)"<< endl;
    all_rz <<"# "<< 16 <<" columns (n,rho,eta,n*vr,n*vt,n*vz,n*kr,n*kt,n*kz,vr,vt,vz,kr-vr^2,kt-vt^2,kz-vz^2,T)"<< endl;
    rho_eta_r <<"# "<< 2 <<" columns (rho, eta)"<< endl;
    rho_eta_rz <<"# "<< 2 <<" columns (rho, eta)"<< endl;
    vr_vt_vz_rz <<"# "<< 3 <<" columns (vr, vt, vz)"<< endl;
    temperature_r <<"# "<< 4 <<" columns (Tr, Tt, Tz, T)"<< endl;
    temperature_rz <<"# "<< 4 <<" columns (Tr, Tt, Tz, T)"<< endl;

// DENSITY NORMALIZATION
    double sum;
    for(int rcell=0; rcell<rcells_; ++rcell) {
        double radius_inv= 1./((rcell + .5)*rscale_inv_);
        for(int zcell=0; zcell<zcells_; ++zcell) {
            int cell= rcell + rcells_*zcell;
            sum+= n_[cell]*radius_inv;
        }
    }
    sum= cells_/sum;

    double hist_scale= 1./entries_;

// ACTUAL DUMPING
    double sphere_vol= shot.diam()*shot.diam()*shot.diam()*M_PI/6.;
    double rzscale_inv= rscale_inv_*zscale_inv_;

    for(int zcell=0; zcell<zcells_; ++zcell) {
        double z= (zcell + .5)*zscale_inv_;
        for(int rcell=0; rcell<rcells_; ++rcell) {
            int cell= rcell + rcells_*zcell;
            double r= (rcell + .5)*rscale_inv_;

            raw_r << r <<' '<< z <<' '<< n_[cell] <<' '
                    << vr_[cell] <<' '<< vt_[cell] <<' '<< vz_[cell] <<' '
                    << kr_[cell] <<' '<< kt_[cell] <<' '<< kz_[cell] << endl;
            raw_rz << r <<' '<< z <<' '<< n_[cell] <<' '
                    << vr_[cell] <<' '<< vt_[cell] <<' '<< vz_[cell] <<' '
                    << kr_[cell] <<' '<< kt_[cell] <<' '<< kz_[cell] << endl;

            double rho=  n_[cell]*sum/r;
            double area_scale= 1/(2*M_PI*r);
            double vol_inv= area_scale*rzscale_inv;
            double eta= hist_scale*n_[cell] *sphere_vol*vol_inv;

            rho_eta_r << r <<' '<< z <<' '<< rho <<' '<< eta <<endl;
            rho_eta_rz << r <<' '<< z <<' '<< rho <<' '<< eta <<endl;

            double cell_scale= 1./n_[cell];
            double
                vr= vr_[cell]*cell_scale,
                vt= vt_[cell]*cell_scale,
                vz= vz_[cell]*cell_scale,
                kr= kr_[cell]*cell_scale,
                kt= kt_[cell]*cell_scale,
                kz= kz_[cell]*cell_scale;
            kr-= vr*vr;
            kt-= vt*vt;
            kz-= vz*vz;

            vr_vt_vz_rz << r <<' '<< z <<' '<< vr <<' '<< vt <<' '<< vz << endl;

            temperature_r << r <<' '<< z <<' '
                    << kr <<' '<< kt <<' '<< kz <<' '<< kr+kt+kz << endl;
            temperature_rz << r <<' '<< z <<' '
                    << kr <<' '<< kt <<' '<< kz <<' '<< kr+kt+kz << endl;

            all_r << r <<' '<< z <<' '<< n_[cell] <<' '<< rho <<' '<< eta <<' '
                    << vr_[cell] <<' '<< vt_[cell] <<' '<< vz_[cell] <<' '
                    << kr_[cell] <<' '<< kt_[cell] <<' '<< kz_[cell] <<' '
                    << vr <<' '<< vt <<' '<< vz <<' '<< kr <<' '<< kt <<' '<< kz <<' '
                    << kr+kt+kz << endl;
            all_rz << r <<' '<< z <<' '<< n_[cell] <<' '<< rho <<' '<< eta <<' '
                    << vr_[cell] <<' '<< vt_[cell] <<' '<< vz_[cell] <<' '
                    << kr_[cell] <<' '<< kt_[cell] <<' '<< kz_[cell] <<' '
                    << vr <<' '<< vt <<' '<< vz <<' '<< kr <<' '<< kt <<' '<< kz <<' '
                    << kr+kt+kz << endl;
        }
        raw_r << endl << endl;
        raw_rz << endl;
        all_r << endl << endl;
        all_rz << endl;
        rho_eta_r << endl << endl;
        rho_eta_rz << endl;
        vr_vt_vz_rz << endl;
        temperature_r << endl << endl;
        temperature_rz << endl;
    }
}

void
Stats3Drz::rescale()
{
}

//------------------------------------------------------------------------------
Stats3Dxyz::~Stats3Dxyz()
{
    dump();
    delete[] hist_;
}

Stats3Dxyz::Stats3Dxyz(const char* filename)
    : entries_(0)
    , hist_(NULL)
{
    {
        ifstream is(filename);
        if(!is.is_open()) {cerr << "Error reading "<< filename << endl; exit(1);}
        double xsize; is >> xcells_ >> xsize;  while(is.get()!='\n');
        xscale_= xcells_/xsize; xscale_inv_= xsize/xcells_;
        double ysize; is >> ycells_ >> ysize;  while(is.get()!='\n');
        yscale_= ycells_/ysize; yscale_inv_= ysize/ycells_;
        double zsize; is >> zcells_ >> zsize;  while(is.get()!='\n');
        zscale_= zcells_/zsize; zscale_inv_= zsize/zcells_;
        if(!is) {cerr << "Error reading "<< filename << endl; exit(1);}
    }
    cells_= xcells_*ycells_*zcells_;
    hist_= new int[cells_];
    for(int cell= 0; cell<cells_; ++cell) hist_[cell]= 0;
}

void
Stats3Dxyz::reset()
{
    for(int cell= 0; cell<cells_; ++cell) hist_[cell]= 0;
    entries_ =0;
}

void
Stats3Dxyz::makeEntry(const Shot& shot)
{
    ++entries_;
    int chunk_size= 3*shot.n();
    const double *r0= shot.r_chunk();
    
    for(int i= 0; i<chunk_size; i+=3) {
        double x= (r0 + i)[X], y= (r0 + i)[Y], z= (r0 + i)[Z];
        int xcell= int(.5*(x*xscale_ + xcells_));
        int ycell= int(.5*(y*yscale_ + ycells_));
        int zcell= int(z*zscale_);
        int cell = xcell + xcells_*ycell + xcells_*ycells_*zcell;
        if(cell>-1 && cell<cells_) {
            ++hist_[cell];
        }
    }
}

void
Stats3Dxyz::dump() const
{
    double hist_scale= 1./entries_;
    const char filename[]= "rho.xyz";
    ofstream os(filename);
    if(!os.is_open()) {cerr << "Error writing "<< filename << endl; exit(1);}
    for(int zcell=0; zcell<zcells_; ++zcell) {
        for(int ycell=0; ycell<ycells_; ++ycell) {
            for(int xcell=0; xcell<xcells_; ++xcell) {
                int cell= xcell + xcells_*ycell + xcells_*ycells_*zcell;
                os << (2.*xcell - xcells_ + 1.)*xscale_inv_ << " "
                   << (2.*ycell - ycells_ + 1.)*yscale_inv_ << " "
                   << (zcell + .5)*zscale_inv_ << " "
                   << hist_[cell]*hist_scale << endl;
            }
            os << endl;
        }
        os << endl;
    }
}
